#!/usr/bin/env python
# -*- coding: utf8
#    Copyright 2013-2015 James S Blachly, MD and The Ohio State University
#
#    This file is part of Mucor.
#
#    Mucor is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Mucor is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Mucor.  If not, see <http://www.gnu.org/licenses/>.

# let print() be a function rather than statement
# ala python3
from __future__ import print_function

# python standard modules
import os
import re
import sys
reload(sys)  
sys.setdefaultencoding('utf8')
import time
import argparse
import csv
import itertools
from collections import defaultdict
import gzip
import cPickle as pickle
import json

# nonstandard modules
import numpy as np
import pandas as pd
from pandas import ExcelWriter
import HTSeq
import pysam

# optional modules

# mucor modules
from config import Config
from mucor_config import parseAndValidateRegions 

def abortWithMessage(message):
    print("*** FATAL ERROR: " + message + " ***")
    exit(2)

def throwWarning(message, help = False):
    print("*** WARNING: " + message + " ***")
    return

def parseJSON(json_config):
    '''
    Import the JSON config file from mucor_config.py. 
    Reads the config file into a dictionary, then writes each dictionary entry into the respective Config class position.
    '''
    config = Config()
    try:
        JD = json.load(open(json_config,'r'))
    except ValueError as json_error:
        throwWarning(json_error.message)
        abortWithMessage("Could not load the given JSON config file. See the example config for proper formatting.")

    # write dictionary values into a more human-friendly config class
    config.outputDir = os.path.expanduser(JD['outputDir'])
    
    # no longer take region definition from json config file
    #   command line argument now
    if str(JD['regions']):
        config.regions = JD['regions']
    else:
        config.regions = []

    config.inputFiles = []
    config.samples = []
    config.bams = {}
    for i in JD['samples']:
        # make a list of bam file paths defined by the JSON config
        #   i.e. extract the path of every file, if the file type is bam
        bams = [x['path'] for x in i['files'] if x['type'] == "bam"]
        if len(bams) >= 1:
            for bam in bams:
                config.inputFiles.append(bam)
                config.samples.append(i['id'])
                try:
                    config.bams[i['id']].append(bam)
                except KeyError:
                    config.bams[i['id']] = []
                    config.bams[i['id']].append(bam)
        else:
            throwWarning("Sample {0} has no BAM file!".format(i['id']))
    if not config.bams:
        abortWithMessage("No bam files defined!")
    return config 

def parseRegionBed(regionfile, regionDictIn):
    ''' 
    Read through the supplied bed file and add the rows to the input region dictionary before returning it. Appends to the input dict, rather than overwriting it.
    '''
    regionDict = regionDictIn
    for line in open(str(regionfile),'r'):
        col = line.strip().split("\t")
        chrom = col[0]
        start = col[1]
        end = col[2]
        if len(col) > 3:
            name = col[3]
        else:
            name = str(chrom) + ":" + str(start) + "-" + str(end)
        regionDict[chrom].add((start,end,name))
    return regionDict

def inRegionDict(chrom, start, end, regionDict):
    '''Checks the given mutation location to see if it is in the dictionary of regions'''
    if regionDict[chrom]: # are there any regions of interest in the same chromosome as this mutation?
        for locs in regionDict[chrom]: # breaking the list of regions according to chromosome should significantly decrease the number of comparisons necessary 
            if locs[0] == 0 and locs[1] == 0: # chrN:0-0 is used to define an entire chromosome as a region of interest. 
                return True
            elif int(start) >= int(locs[0]) and int(end) <= int(locs[1]):
                return True
    return False

def GaugeDepth(config) :
    '''
    assess the depth of each sample at the given regions
    '''

    startTime = time.clock()
    regionDict = defaultdict(set)
    for item in config.regions:
        if str(str(item).split('.')[-1]).lower() == 'bed':      # this item is a bed file
            regionDict = parseRegionBed(item, regionDict)
        else:                                                   # this was defined as a string
            reg_chr = item[0]
            reg_str = item[1]
            reg_end  = item[2]
            if reg_str >= 0 and reg_end > reg_str:
                name = str(reg_chr) + ":" + str(reg_str) + "-" + str(reg_end)
            elif reg_str >= 0 and reg_str == reg_end:
                name = str(reg_chr) + ":" + str(reg_str)
            elif not reg_str and not reg_end:
                name = str(reg_chr)
            regionDict[reg_chr].add((reg_str, reg_end, name))

    if not regionDict:
        abortWithMessage("Regions not set!")

    covD = {'chr':[], 'start':[], 'stop':[], 'name':[], 'sample':[],'depth':[]}
    print("\n=== Reading BAM Files ===")
    for sid, fns in config.bams.items():
        # loop over all samples
        for fn in fns:
            # loop over all bam files for this sample
            try:
                samfile = pysam.AlignmentFile(fn, "rb" )
            except ValueError:
                throwWarning("Cannot open file {0}".format(fn))
                continue
            contig_length_dict = dict(zip(samfile.references, samfile.lengths)) # save contig lengths for later
            for contig, ROI in regionDict.items():
                for window in ROI:
                    bed_name = window[2]
                    # make window 0-based
                    try:
                        window = [int(window[0]) - 1, int(window[1])]
                    except TypeError: 
                        # start and/or stop were not defined - pull them from contig dictionary
                        if not window[0]:
                            start = 0
                        else:
                            start = int(window[0] - 1)
                        if not window[1]:
                            end = contig_length_dict[contig]
                        else:
                            end = int(window[1])
                        window = [start, end]
                    # loop over all ROIs, checking this bam 
                    if config.p:
                        #point method 
                        tmp_dict = {}
                        position = round((window[1] - window[0])/2.0) + window[0]
                        avg_covg = samfile.count(contig, position - 1, position)

                        #for position in range(window[0],window[1]):
                        #    region = str(contig) + ':' + str(position) + '-' + str(position)
                        #    tmp_dict[position] = samfile.count(region=region)
                        #avg_covg = np.mean(tmp_dict.values())
                    elif config.c:
                        #read count method
                        avg_covg = samfile.count(contig, window[0], window[1])
                        #note that "avg_covg" is only a name here - it is the total count of reads, not an average! 
                    else:
                        #complete average method
                        #'''
                        tmp_dict = {}
                        for position in range(window[0],window[1]):
                            tmp_dict[position] = 0
                        for pileupcolumn in samfile.pileup(contig, window[0], window[1],stepper='all'):
                            #loop over reads that hit the window locations and record coverage 
                            # 'stepper = all' yields mapped, primary, non-duplicate (identified by sam flag), QC pass reads
                            try:
                                tmp_dict[pileupcolumn.pos]
                                tmp_dict[pileupcolumn.pos] = pileupcolumn.n       
                            except KeyError:
                                #skip this position if it's not in the region dict
                                continue
                        avg_covg = np.mean(tmp_dict.values())
                        #'''
                        '''
                        #this behaves erratically and does not produce the same number if run repeatedly
                        #   not sure how to use the function, but it could be faster than pileup
                        counter = 0
                        for ct_cov in samfile.count_coverage(contig, window[0], window[1], read_callback = 'all'):
                            for nt_arr in ct_cov:
                                counter += int(nt_arr)
                        stop()
                        avg_covg = counter/float(window[1] - window[0])
                        '''


                    covD['chr'].append(str(contig))
                    covD['start'].append(int(window[0]) + 1)
                    covD['stop'].append(int(window[1]))
                    covD['name'].append(str(bed_name))
                    covD['sample'].append(str(sid))
                    covD['depth'].append(float(avg_covg))     
            samfile.close()
            totalTime = time.clock() - startTime
            print("{0:02d}:{1:02d}\t{2}".format(int(totalTime/60), int(totalTime % 60), fn))
    covDF = pd.DataFrame.from_dict(covD)[['chr','start','stop','name','sample','depth']]
    covDF = covDF.groupby(['chr','start','stop','name','sample'])['depth'].apply(sum).reset_index()
            
    totalTime = time.clock() - startTime
    print("\n{0:02d}:{1:02d}\t{2}".format(int(totalTime/60), int(totalTime % 60), "Done"))
    return covDF

def SamplesToColumns(grp):
    '''helper function to rearrange the coverage dataframe'''
    columns = ['chr','start','stop','name']
    values = [grp['chr'].values[0], grp['start'].values[0], grp['stop'].values[0],grp['name'].values[0]]

    grp_dict = {}
    for i in grp[['sample','depth']].T.itertuples(): 
        grp_dict[i[0]] = list(i[1:])
    columns += grp_dict['sample']
    values += grp_dict['depth']
    return pd.DataFrame.from_dict(dict(zip(columns,values), index=[0]))[columns]


def printOutput(config, covDF):
    '''reformat the coverage dataframe and write it to xlsx file'''

    startTime = time.clock()
    print("\n=== Writing output files to {0}/ ===".format(config.outputDir))
    outDF = covDF.groupby(['chr','start','stop','name']).apply(SamplesToColumns).reset_index(drop=True)
    # pandas can actually start up an excel writer object using a non-existant or unwritable file path. The error will come when saving to such a path. tested in pandas version 0.16.2
    outXLSX = pd.ExcelWriter(str(config.outputDir) + "/" + config.outfile, engine='xlsxwriter')
    outDF.to_excel(outXLSX, 'Depth Gauge', index=False)
    try:
        outXLSX.save()
    except IOError:
        abortWithMessage("Output directory is not writable or doesn't exist: {0}".format(config.outputDir))
    totalTime = time.clock() - startTime
    print("\t{0}: {1} rows".format(str(config.outputDir) + "/" + config.outfile , len(outDF)))     
    return True

def main():
    print("\n=== Run Info ===")
    print("\t{0}".format(time.ctime() ) )
    print()

    parser = argparse.ArgumentParser()
    parser.add_argument("json", help="Pass 1 json config as an argument")
    parser.add_argument("-p", "--point", dest='p', action="store_true", help="point-location coverage approximation on middle of region windows")
    parser.add_argument("-c", "--count", dest='c', action="store_true", help="report total number of reads in each region window (not an average)")
    parser.add_argument("-r", "--regions", default="", help="Comma separated list of bed regions and/or bed files by which to limit output. Ex: chr1:10230-10240,chr2,my_regions.bed")
    args = parser.parse_args()
    config = parseJSON(args.json)
    config.p = bool(args.p)
    config.c = bool(args.c)
    if config.p and config.c:
        abortWithMessage("Cannot declare -p and -c options together!")
    if config.p:
        config.outfile = 'Depth_of_Coverage-p.xlsx'
    elif config.c:
        config.outfile = 'Depth_of_Coverage-c.xlsx'
    else:
        config.outfile = 'Depth_of_Coverage.xlsx'
    if os.path.exists(config.outputDir) and config.outfile in [str(x) for x in os.listdir(config.outputDir)]:
        abortWithMessage("The directory {0} already exists and contains depth gauge output. Will not overwrite.".format(config.outputDir))
    elif not os.path.exists(config.outputDir):
        try:
            os.makedirs(config.outputDir)
        except OSError:
            abortWithMessage("Error when creating output directory {0}".format(config.outputDir))

    # check that all specified variant files exist
    for fn in config.inputFiles:
        if not os.path.exists(fn):
            abortWithMessage("Could not find variant file: {0}".format(fn))
    config.regions = parseAndValidateRegions(args.regions, config) 
    covDF = GaugeDepth(config)
    printOutput(config, covDF)
    # pretty print newline before exit
    print()


if __name__ == "__main__":
    if sys.hexversion < 0x02070000:
        raise RuntimeWarning("mucor should be run on python 2.7.0 or greater.")
    main()



#!/usr/bin/env python
# -*- coding: utf8
#    Copyright 2013-2015 James S Blachly, MD and The Ohio State University
#
#    This file is part of Mucor.
#
#    Mucor is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Mucor is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Mucor.  If not, see <http://www.gnu.org/licenses/>.

# mucor.py
# (c) James S Blachly, MD 2013
# 

# let print() be a function rather than statement
# ala python3
from __future__ import print_function

# python standard modules
import os
import sys
import time
import argparse
import csv
import itertools
from collections import defaultdict
import gzip
import cPickle as pickle
import json

# nonstandard, required modules
try:
    import numpy as np
except ImportError as err:
    print(str(err) + ". This module is required.")
    sys.exit()
try:
    import pandas as pd
except ImportError as err:
    print(str(err) + ". This module is required.")
    sys.exit()
try:
    import HTSeq
except ImportError as err:
    print(str(err) + ". This module is required.")
    print("* NOTE! If you installed with pip install mucor,")
    print("* the numpy and pandas dependencies were automatically")
    print("* installed. However, due to a bug in the HTSeq")
    print("* installation, it cannot be automatically installed")
    print("* at the same time as numpy.")
    print("* SOLUTION: pip install HTSeq")
    sys.exit()

# optional modules
# will throw warnings later in the program 
try:
    import tabix
except ImportError:
    pass

try:
    import xlsxwriter
except ImportError:
    pass

# mucor modules
import mucorfilters as mf
from variant import Variant
from mucorfeature import MucorFeature
import inputs
import output
from config import Config
from databases import dbLookup,checkAndOpen 
from info import Info

def abortWithMessage(message):
    print("*** FATAL ERROR: " + message + " ***")
    exit(2)

def throwWarning(message, help = False):
    print("*** WARNING: " + message + " ***")
    return

def constructGAS(gffFile, featureType, knownFeatures, duplicateFeatures, union):
    gas = HTSeq.GenomicArrayOfSets("auto", stranded=False)
    for feature in itertools.islice(gffFile, 0, None):
        # Nonstandard contigs (eg chr17_ctg5_hap1, chr19_gl000209_random, chrUn_...)
        # must be specifically excluded, otherwise you will end up with exception
        # ValueError: start is larger than end
        # due to duplicated gene symbols
        if "_hap" in feature.iv.chrom:
            continue
        elif "_random" in feature.iv.chrom:
            continue
        elif "chrUn_" in feature.iv.chrom:
            continue

        # skip full transcripts, full genes, and other items that are not exon ranges
        # including these lines will create '--union'-like behavior, regardless of whether the user passed the union option
        if feature.type != 'exon':
            continue

        # transform feature to instance of Class MucorFeature
        feat = MucorFeature(feature.attr[featureType], feature.type, feature.iv)
        
        # WARNING
        # the following REQUIRES a coordinate-sorted GFF/GTF file
        # extra checks incurring slowdown penalty are req'd if GFF/GTF not sorted

        if feat.name in knownFeatures:
            # In case there is an error in the GFF and/or the featureType (-f) is not unique,
            # issue a warning
            # for example, genes.gtf supplied with the Illumina igenomes package for the tuxedo tools suite
            # includes duplicate entries for many genes -- e.g. DDX11L1 on chr15 shoudl be DDX11L9
            # try to cope with this by relabeling subsequent genes as GENESYM.chrNN
            if feat.iv.chrom != knownFeatures[feat.name].iv.chrom:
                duplicateFeatures.add(feat.name)
                feat.name = feat.name + '.' + feat.iv.chrom
            else:
                if union:
                    # replace the start and end coordinates when adding SUCCESSIVE bits of a feature (e.g. exons)
                    # results in one, large feature, starting with the earliest upsream location and ending with the latest downstream location
                    if knownFeatures[feat.name].iv.start < feat.iv.start:
                        feat.iv.start = knownFeatures[feat.name].iv.start
                    if knownFeatures[feat.name].iv.end > feat.iv.end:
                        feat.iv.end = knownFeatures[feat.name].iv.end
                else:
                    pass # no-union - this does overwrite previous coordinates in knownFeatures,
                         # but should not matter as the actual coordinates are obtaind from 'gas'.
                         # Locations held in knownFeatures should not be used to identify the start and stop of a feature,
                         # since these locations will only reveal the last known region for the feature. 
                         # Later in the program, querying knownFeatures for the variants in a feature (ie: skipThisIndel function)
                         # may return variants that appear to be located in positions outside of the feature region reported by knownFeatures. 
                         # This is also caused by the 'no-union' overwrite, since the knownFeature region locations do not represent the whole feature.

        # first, add to the knownFeatures, a dictionary of MucorFeatures, which contain the variants set
        knownFeatures[feat.name] = feat
        # then, add to the GenomicArrayOfSets, which we use to find gene symbol from variant coords
        try:
            gas[ feat.iv ] += feat.name
        except ValueError:
            print(feat.name)
            print(feat.iv)
            raise
    
    return gas, knownFeatures, duplicateFeatures

def parseJSON(json_config):
    '''
    Import the JSON config file from mucor_config.py. 
    Reads the config file into a dictionary, then writes each dictionary entry into the respective Config class position.
    '''
    config = Config()
    try:
        JD = json.load(open(json_config,'r'))
    except ValueError as json_error:
        throwWarning(json_error.message)
        abortWithMessage("Could not load the given JSON config file. See the example config for proper formatting.")

    # write dictionary values into a more human-friendly config class
    config.featureType = JD['feature']
    config.outputDir = os.path.expanduser(JD['outputDir'])
    config.union = JD['union']
    config.fast = JD['fast']
    config.gff = JD['gff']
    config.outputFormats = list(set(JD['outputFormats'])) # 'set' prevents repeated formats from being written multiple times
    if JD['databases']:
        if 'tabix' in sys.modules: # make sure tabix is imported 
            for name,db in JD['databases'].items():
                dbPointer = checkAndOpen(db)
                if dbPointer:
                    #check for non-null pointers
                    config.databases[name] = dbPointer
        else: # the user supplied databases but did not successfully import tabix
            throwWarning("tabix module not found; database features disabled")
    if str(JD['regions']):
        config.regions = JD['regions']
    else:
        config.regions = []
    # comma separated list of acceptable VCF filter column values
    config.filters = JD['filters']

    config.inputFiles = []
    config.samples = []
    for i in JD['samples']:
        config.samples.append(i['id'])
        for j in i['files']:
            if j['type'] == "bam":
                # skip the bam files
                continue
            filename = os.path.basename(j['path'])
            config.filename2samples[filename] = i['id']
            config.source[filename] = j['source']
            config.inputFiles.append(j['path'])

    return config 

def parseGffFile(gffFileName, featureType, fast, union):
    '''
    Parse the GFF/GTF file. Return tuple (knownFeatures, GenomicArrayOfSets)
    Haplotype contigs are explicitly excluded because of a coordinate crash (begin > end)
    '''
    
    # TO DO: command line flag should indicate that variants in INTRONS are counted
    # This is called --union, see below
    
    startTime = time.clock()
    print("\n=== Reading GFF/GTF file {0} ===".format(gffFileName))
    
    if union:
        unionstatus = str("union")
    else:
        unionstatus = str("no_union")
    annotFileName = os.path.splitext(os.path.basename(gffFileName))[0]
    # pickled file is created for specific combinations of gff annotation and feature type. See below for more details. **

    archiveFilePath = os.path.expanduser(fast) + str("/") + str(annotFileName) + str('_') + str(featureType) + str("_") + unionstatus + str('.p')
    print(gffFileName)
    gffFile = HTSeq.GFF_Reader(gffFileName)
    
    knownFeatures = {}

    duplicateFeatures = set()

    # gas - GenomicArrayOfSets.
    # typecode always 'O' (object) for GenomicArrayOfSets
    # UNstranded -- VCF and muTect output always report on + strand,
    # but the GenomicArray must be unstranded because the GFF /is/ strand-specific,
    # and if I manually coded all GenomicIntervals read from the VCF or muTect file as '+',
    # then no genes on the - strand would have variants binned to them

    # Fast boolean determines whether the user wants to load a pickle file annotation that they made in a previous run
    # The pickle file will load the genomic array of sets, known features, and duplicate features. 
    # ** These three items will change depending on the supplied gff annotation AND the feature selected AND whether union was used. 
    #    Thus, each pickle file will be for a specific combination of gff, feature, and union status, hence the naming convention.
    if bool(fast):
        # user wants to use pickled annotations
        if os.path.exists( archiveFilePath ):
            # using the existing, pickled annotation
            print("Opening annotation archive: " + str(archiveFilePath))
            pAnnot = open(archiveFilePath, 'rb')
            gas = pickle.load(pAnnot)
            knownFeatures = pickle.load(pAnnot)
            duplicateFeatures = pickle.load(pAnnot)
            pAnnot.close()
        else:
            # no pickled annotation exists for this combination of gff and feature; creating it in the directory provided to the 'fast' option
            print("Cannot locate annotation archive for " + os.path.basename(gffFileName) + str(" w/ ") + str(featureType) + " and --union=" + str(union) )
            print("   Reading in annotation and saving archive for faster future runs") 
            if not os.path.exists( str("/".join(archiveFilePath.split('/')[:-1])) ):
                os.makedirs( str("/".join(archiveFilePath.split('/')[:-1])) )
            gas, knownFeatures, duplicateFeatures = constructGAS(gffFile, featureType, knownFeatures, duplicateFeatures, union)
            archiveOut = open(archiveFilePath, 'wb')
            pickle.dump(gas, archiveOut, -1)
            pickle.dump(knownFeatures, archiveOut, -1)
            pickle.dump(duplicateFeatures, archiveOut, -1)
            archiveOut.close()
    if not bool(fast):
    # ignore pickles function entirely. Won't check for it and won't attempt to create it
        gas, knownFeatures, duplicateFeatures = constructGAS(gffFile, featureType, knownFeatures, duplicateFeatures, union)

    if duplicateFeatures:
        print("*** WARNING: {0} {1}s found on more than one contig".format(len(duplicateFeatures), featureType))
    # remove gene-centric feature names that are known to have problematic bins caused by multiple copies on the same gene
    # run the accompanying 'detect union bin errors' python script to generate a list of incompatible genes
    if union and "gene" in featureType:
        try:
            badGenes = open('union_incompatible_genes.txt')
            badGeneList = set()
            print("Removing genes with multiple copies on the same contig, which cause incorrect feature bins with 'union'\n\tUsing 'union_incompatible_genes.txt'")
            for line in badGenes:
                badGene = line.strip()
                badGeneList.add(badGene)
            for interval, sets in gas.steps():
                gas[interval] = set( [ x for x in sets if x not in badGeneList ] )
            for badGene in badGeneList:
                try:
                    del knownFeatures[badGene]
                    # uncomment below to see which genes are being deleted 
                    # print(badGene + " deleted from knownfeat")
                except KeyError:
                    pass
        except IOError:
            # There likely isn't a list of genes
            # Recommend that the user obtain said list, or face potential errors in feature labels and counts
            pass
    else:
        # not using a gene_name-type of feature in combination with --union
        # no known issues should arise with this configuration 
        pass

    totalTime = time.clock() - startTime
    print("{0} sec\t{1} found:\t{2}".format(int(totalTime), featureType, len(knownFeatures)))

    return knownFeatures, gas

def parseRegionBed(regionfile, regionDictIn):
    ''' 
    Read through the supplied bed file and add the rows to the input region dictionary before returning it. Appends to the input dict, rather than overwriting it.
    '''
    regionDict = regionDictIn
    for line in open(str(regionfile),'r'):
        col = line.split("\t")
        chrom = col[0]
        start = col[1]
        end = col[2]
        regionDict[chrom].add((start,end))
    return regionDict

def filterRow(row, fieldId, filters, kind):
    '''
    returning True means this row will be filtered out [masked]
    returning False means the row will not be filtered out [pass filter]
    '''
    if kind in ["vcf", "vcf.gz"]:
        # this is handled in 1 line elsewhere in the program. Specifically, within the parseVariantFiles function, under the [vcf, vcf.gz] block
        for rowFilter in str(row[fieldId['FILTER']]).split(';'):    ## VCF file format
            if rowFilter not in filters:
                return True
                break
    if str(kind) == "out":
        for rowFilter in str(row[fieldId['judgement']]).split(';'): ## MuTect '.out' file format
            if rowFilter not in filters:
                return True
                break
    return False

def inRegionDict(chrom, start, end, regionDict):
    '''Checks the given mutation location to see if it is in the dictionary of regions'''
    if regionDict[chrom]: # are there any regions of interest in the same chromosome as this mutation?
        for locs in regionDict[chrom]: # breaking the list of regions according to chromosome should significantly decrease the number of comparisons necessary 
            if locs[0] == 0 and locs[1] == 0: # chrN:0-0 is used to define an entire chromosome as a region of interest. 
                return True
            elif int(start) >= int(locs[0]) and int(end) <= int(locs[1]):
                return True
    return False

def skipThisIndel(var, knownFeatures, featureName):
    '''
    Input a mutation and the current list of known mutations and features
    If this mutation already exists in another form, returns True, along with the existing ref and alt.
        Existing in another form is defined as 2 mutations that occur in the same position and have different ref/alt, but are functionally equivalent
        Ex: ref/alt: A/AT compared to ref/alt: ATT/ATTT
        The refs and the alts are different, but both mutations represent the same T insertion
    '''
    
    for kvar in knownFeatures[featureName].variants: # "kvar" stands for "known variant"
        if kvar.pos.pos == var.pos.pos and kvar.pos.chrom == var.pos.chrom and \
        ( kvar.ref != var.ref or kvar.alt != var.alt ) and \
        indelDelta(var.ref,var.alt)[0] == indelDelta(kvar.ref, kvar.alt)[0] and \
        indelDelta(var.ref,var.alt)[1] == indelDelta(kvar.ref, kvar.alt)[1]:
            return (kvar.ref, kvar.alt)
            break
    return tuple()

def indelDelta(ref, alt):
    ''' Detects the inserted or deleted bases of an indel'''
    if len(alt) > len(ref):             # insertion
        return [ alt.replace(ref,"",1), str("INS") ]
    elif len(alt) < len(ref):           # deletion
        return [ ref.replace(alt,"",1), str("DEL") ]
    else:                               # SNP / MNP
        # there is already a SNP at this indel location 
        return [ "", "" ]

def groupCount(grp):
    ''' 
    function to count the number of samples that share the same mutation
    Input: Acts on a pandas groupby object
    Oputput: returns a DataFrame
    DEPRICATED: This has been reduced to a faster lambda function 
    '''
    return len(grp['sample'].unique())

def annotateDF(grp, databases):
    '''
    function to annotate variant dataframe with user-supplied vcf databases
    '''
    chrom = grp['chr'].unique()[0]
    pos = grp['pos'].unique()[0]
    ref = grp['ref'].unique()[0]
    alt = grp['alt'].unique()[0]
    var = Variant(source=None, sample=None, pos=HTSeq.GenomicPosition(chrom, pos), ref=ref, alt=alt, frac=None, dp=None, eff=None, fc=None)
    dbEntries = dbLookup(var, databases)
    return pd.Series(dbEntries)

def integrateVar(var, varD, config, gas, knownFeatures, unrecognizedContigs, unrecognizedMutations):
    '''
    Ingest the given variant object into the variant dictionary 
    '''
    # find bin for variant location
    try:
        resultSet = gas[ var.pos ]      # returns a set of zero to n IDs (e.g. gene symbols)
    except KeyError:                    # which I'll use as a key on the knownFeatures dict
                                        # and each feature with matching ID gets allocated the variant
        # this mutation is on a contig unknown to the GAS
        resultSet = set()
        unrecognizedContigs.add(var.pos.chrom)
        unrecognizedMutations += 1
   
    if resultSet:
        for featureName in resultSet:
            if len(var.ref) != len(var.alt): # confirm that this mutation is an indel, not a snp
                kvar = skipThisIndel(var, knownFeatures, featureName)
                if bool(kvar):
                    # Sanity check to see what indels are being overwritten by existing vars
                    var.ref = kvar[0]
                    var.alt = kvar[1]
                knownFeatures[featureName].variants.add(var)
    
    # Descriptive variable names
    chr = var.pos.chrom
    pos = int(var.pos.pos)
    ref = var.ref
    alt = var.alt 
    vf = var.frac
    dp = var.dp
    features = ', '.join( resultSet )   # join with comma to handle overlapping features
    effect = var.eff
    fc = var.fc
    sample = var.sample
    source = var.source
    count = 0 # initialize this database column now to save for later
    freq = 0.0 # initialize this database column now to save for later
    
    # build dict to insert
    # Step 1: define the columns and their values
    # Step 2. zip the column names and column values together into a dictionary
    # Step 3. Add this round to the master variants data frame dictinoary

    for feature in features.split(', '):
        # Removing columns from the following 'columns' list will mask them from output
        columns = ['chr','pos','ref','alt','vf','dp','feature','effect','fc','count','freq','sample','source']
        values  = [ chr,  pos,  ref,  alt,  vf,  dp,  feature,  effect,  fc , count,  freq,  sample, source  ]

        vardata = dict(zip( columns, values ))
        for key in vardata.keys():
            varD[key].append(vardata[key])
    return varD, unrecognizedContigs, unrecognizedMutations

def parseVariantFiles(config, knownFeatures, gas, databases, filters, regions, total) :
    '''
    Read in all input files
    Record mutations in the variant dataframe 
    '''

    startTime = time.clock()
    unrecognizedContigs = set()
    unrecognizedMutations = 0

    # All variants stored in long (record) format
    # in a pandas dataframe
    #
    # However, it is MUCH faster to initially store them in a python Dict
    # Then convert to the pandas DF at the end
    varD = defaultdict(list) 
    variantFiles = set(list(config.inputFiles)) # uniquify the file list, in the case of the same multi-sample VCF being defined for multiple samples

    if regions: # has the user specified any particular regions or region files to focus on?
        regionDict = defaultdict(set)
        for item in regions:
            if os.path.splitext(item)[1].lower() == 'bed':      # this item is a bed file
                regionDict = parseRegionBed(item, regionDict)
            elif str(str(item).split(':')[0]).startswith('chr'):    # this is a string 
                reg_chr = str(item.split(':')[0])
                try:
                    reg_str = str(str(item.split(':')[1]).split('-')[0])
                    reg_end = str(str(item.split(':')[1]).split('-')[1])
                except IndexError:                                  # represent whole chromosome regions [ex: chr2] by chrN:0-0 in the region dictionary   
                    reg_str = 0
                    reg_end = 0
                regionDict[reg_chr].add((reg_str, reg_end))

    print("\n=== Reading Variant Files ===")
    for fn in variantFiles:
        kind = str( os.path.splitext(fn)[-1].strip('.').lower() )
        if kind == "gz" and fn.endswith(".vcf.gz"):
            kind = "vcf.gz"
        try:
            varFile = open(fn, 'rb')
        except IOError:
            # file does not exist or cannot be opened
            throwWarning('{} could not be opened'.format(fn))
            continue    # next fn in variantFiles

        if kind == "out":
            # parse as mutect '.out' type format
            varReader = csv.reader(varFile, delimiter="\t")
            row = varReader.next()
            while row[0].startswith('##'):
                row = varReader.next()
            header = row
            if len(header) == 0: raise ValueError('Invalid header')
            fieldId = dict(zip(header, range(0, len(header))))
            for row in itertools.islice(varReader, None):
                if filterRow(row, fieldId, filters, kind):  # filter rows as they come in, to prevent them from entering the dataframe
                    continue                                # this allows us to print the dataframe directly and have consistent output with variant_details.txt, etc.
                source = config.source[ os.path.basename(fn) ]
                parser = inputs.Parser()
                parser.row = row
                parser.source = source
                parser.fieldId = fieldId
                parser.header = header
                parser.fn = fn 
                var = parser.parse_MuTectOUT()
                var.sample = config.filename2samples[os.path.basename(fn)]
                if not var:
                    # this mutation had no data in this sample
                    continue
                if regions and not inRegionDict(var.pos.chrom, int(var.pos.pos), int(var.pos.pos), regionDict ):
                    continue
                varD, unrecognizedContigs, unrecognizedMutations = integrateVar(var, varD, config, gas, knownFeatures, unrecognizedContigs, unrecognizedMutations)

        elif kind in ["vcf", "vcf.gz"]:
            # start htseq vcf reader
            varReader = HTSeq.VCF_Reader(str(fn))
            varReader.parse_meta()
            varReader.make_info_dict()
            for row in varReader:
                if row.filter not in filters:
                    continue
                if regions and not inRegionDict(row.pos.chrom, int(row.pos.pos), int(row.pos.pos), regionDict ):
                    continue
                row.unpack_info(varReader.infodict)
                parser = inputs.Parser()
                source = config.source[ os.path.basename(fn) ]
                try:
                    samps = parser.parse(row, source)
                except KeyError:
                    throwWarning("parsing " + fn + " as '" + source + "'")
                    print("Cannot parse file from an unsupported or unknown variant caller. \nPlease use supported variant software, or compose an input module compatible with inputs.py")
                    print("Options include: " + str(parser.supported_formats.keys()).replace("'",""))
                    break
                for var in samps:
                    if len(samps) == 1:
                        # if not multi-sample VCF, pull sample ID from the json config
                        var.sample = config.filename2samples[ os.path.basename(fn) ]
                    elif len(samps) > 1:
                        # if multi-sample VCF, use sample ID as defined by the VCF column(s) rather than the config
                        pass
                    if not var or var.sample not in config.samples:
                        # this sample has no mutation data at the given location, or this sample was not specified as a sample of interest in the JSON config 
                        continue
                    var.source = os.path.basename(fn)
                    
                    varD, unrecognizedContigs, unrecognizedMutations = integrateVar(var, varD, config, gas, knownFeatures, unrecognizedContigs, unrecognizedMutations)
        else:
            throwWarning("Unable to parse file with extension '{0}': {1}".format(kind, fn))
            continue
        if unrecognizedContigs:
            throwWarning("{0} Contigs and {1} mutations are in areas unknown to the genomic array of sets. If using --archive_directory, perhaps try again without it.".format( len(unrecognizedContigs), unrecognizedMutations ))
        totalTime = time.clock() - startTime
        print("{0:02d}:{1:02d}\t{2}".format(int(totalTime/60), int(totalTime % 60), fn))
    
    columns = ['chr','pos','ref','alt','vf','dp','feature','effect','fc','count','freq','sample','source']
    # Transform data frame dictionary into pandas DF. Major speed increase compared to appending variants to the DF while reading the input files. 
    try:
        varDF = pd.DataFrame(varD, columns=columns)
        # Drop samples with no detected mutation at the given loc
        varDF = varDF.dropna(subset=['dp','vf'], how='all')
        # Cast dp, vf, and pos as appropriate types
        varDF['dp'] = varDF['dp'].astype(int)
        varDF['vf'] = varDF['vf'].astype(float)
        varDF.pos   = varDF.pos.astype(int)
    except UnboundLocalError:
        if not varD:            
            # variant dictionary is empty, implying no variants were encountered and kept
            abortWithMessage("Variant DataFrame is empty! No mutations passed filter; check FILTER column of input VCF against allowed filters in JSON config.")
        else:
            abortWithMessage("Could not convert Variant Dictionary into a DataFrame")
    except ValueError:
        # likely couldn't convert NaN values from dp column to int
        # proceed anyway, treating them as floats instead
        varDF['dp'] = varDF['dp'].astype(float)
    if len(varDF) == 0:
        abortWithMessage("Variant DataFrame is empty! No mutations passed filter; check FILTER column of input VCF against allowed filters in JSON config.")
    # Annotate dataframe using user-supplied vcf databases
    if databases:
        print("\n=== Comparing Your Variants to Known VCF Databases ===")
        startTime = time.clock()
        annotSeries = varDF.groupby(['chr', 'pos', 'ref', 'alt']).apply( annotateDF, databases=databases )
        annotDF = pd.DataFrame(annotSeries)
        new_index = pd.Index([x for x in varDF.columns[:-4]] + [x for x in annotDF.columns] + [x for x in varDF.columns[-4:]])
        varDF = varDF.merge(annotDF, left_on=["chr", "pos", "ref", "alt"], right_index=True)[new_index]
        totalTime = time.clock() - startTime
        print("{0:02d}:{1:02d}".format(int(totalTime/60), int(totalTime % 60)))

    # Delete VAF columns from databases that do not have population variant allele frequencies recorded
    # These will manifest as a whole VAF column full of '?' 
    for column in varDF.keys():
        if "_VAF" in str(column) and len(set(varDF[column])) == 1:
            varDF.drop(column, axis=1, inplace=True)
 
    # Dataframe operation to count the number of samples that exhibit each mutation
    countSeries = varDF.groupby(['chr', 'pos', 'alt']).apply(lambda x: len(set(x['sample'])))
    countDF = pd.DataFrame(countSeries, columns=['count'])
    # This appears messy, but runs very fast regardless. It maps the grouped count dataframe back to the variant dataframe 
    varDF['count'] = varDF[['chr','pos','alt']].merge(countDF, left_on=['chr','pos','alt'], right_index=True)['count']
    # Divide the count for each row in varDF by the total sample count from the JSON config file. Represents the percent of the input samples that exhibit each mutation.
    freqdict = {}
    for i in set(varDF['count'].values):
        freqdict[i] = float(i)/float(total)
    varDF['freq'] = varDF['count'].map(freqdict)

    # blank features should be more descriptive. this replaces empty strings with 'NO_FEATURE'. 
    #   this is important to later functions that put feature in a pandas index. na_rep pandas function only changes dataframe values, not the index.
    #   see output.FeatureXSample() for an example of when 'feature' is used in a pandas index.
    
    # Avoid chained indexing:
    # use DataFrame.loc[row_index, col_index]
    varDF.loc[ varDF.feature == '', 'feature'] = 'NO_FEATURE'
    
    # replace all remaining blank cells with '?'

    #stop() # this command throws a warning
    varDF.replace('', '?', inplace=True)

    return varDF, knownFeatures, gas 

def printOutput(config, outputDirName, varDF):
    '''Output run statistics and variant details to the specified output directory.'''

    startTime = time.clock()
    print("\n=== Writing output files to {0}/ ===".format(outputDirName))
    ow = output.Writer()
    # identify formats ending in 'xlsx' as the "excel formats," requiring XlsxWriter
    excel_formats = [ plugin_name for plugin_name,file_name in ow.file_names.items() if os.path.splitext(file_name)[1]==".xlsx" ]
    if "all" in config.outputFormats:
        # replace output type 'all' with a lsit of all supported output types
        #    and remove 'all' and 'default' to prevent recursive execution of the modules
        config.outputFormats = ow.supported_formats.keys()
        config.outputFormats.remove('all')
        config.outputFormats.remove('default')
        if 'xlsxwriter' not in sys.modules and excel_formats:
            # if xlsxwriter is not present and the user selected excel output formats, remove excel formats from output formats
            config.outputFormats = [ x for x in config.outputFormats if x not in excel_formats ]
            throwWarning("xlsxwriter module not found; Excel outputs disabled")

    for format in config.outputFormats:
        ow.write(varDF,format,outputDirName,config)

    totalTime = time.clock() - startTime
    print("\tTime to write: {0:02d}:{1:02d}".format(int(totalTime/60), int(totalTime % 60)))
    return 0

def main():
    print(Info.logo)
    print(Info.versionInfo)

    print("\n=== Run Info ===")
    print("\t{0}".format(time.ctime() ) )
    print()

    parser = argparse.ArgumentParser()
    parser.add_argument("json", help="Pass 1 json config as an argument")
    args = parser.parse_args()
    config = parseJSON(args.json)

    if not os.path.exists(config.gff):
        abortWithMessage("Could not find GFF file {0}".format(config.gff))
    if os.path.exists(config.outputDir) and [x for x in os.listdir(config.outputDir) if x in output.Writer().file_names.values() ]:
        abortWithMessage("The directory {0} already exists and contains output. Will not overwrite.".format(config.outputDir))
    elif not os.path.exists(config.outputDir):
        try:
            os.makedirs(config.outputDir)
        except OSError:
            abortWithMessage("Error when creating output directory {0}".format(config.outputDir))

    # check that all specified variant files exist
    for fn in config.inputFiles:
        if not os.path.exists(fn):
            abortWithMessage("Could not find variant file: {0}".format(fn))

    # Total is used in frequency calculations. 
    #   supplying config.samples will use sample count as the denominator [canonical operation, ie: comparing samples]
    #   or, using samples.inputFiles will use file count [non-canonical operation, ie: comparing tools, or otherwise comparing many vcf files with no regard for sample ID]
    total = len(set(config.samples))

    knownFeatures, gas = parseGffFile(str(config.gff), str(config.featureType), config.fast, config.union)
    varDF, knownFeatures, gas = parseVariantFiles(config, knownFeatures, gas, config.databases, config.filters, config.regions, total)
    printOutput(config, str(config.outputDir), varDF)
    
    # pretty print newline before exit
    print()


if __name__ == "__main__":
    if sys.hexversion < 0x02070000:
        raise RuntimeWarning("mucor should be run on python 2.7.0 or greater.")
    main()



from setuptools import setup, find_packages

setup(name="mucor",
    version="1.51",
    description="Genomic Variant Aggregation and Mutation Correlation",
    author="James Blachly",
    author_email="james blachly at gmail com",
    license="GNU GPLv3",
    url="http://github.com/blachlylab/mucor",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["pandas", "numpy", "pytabix", "xlsxwriter"],
    scripts=['mucor.py', 'mucor_config.py', 'depth_gauge.py'],
    keywords = "sequencing, VCF, mutation, variant",
    classifiers = [ "Development Status :: 5 - Production/Stable",
                    "Environment :: Console",
                    "Intended Audience :: Healthcare Industry",
                    "Intended Audience :: Science/Research",
                    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                    "Natural Language :: English",
                    "Operating System :: OS Independent",
                    "Programming Language :: Python",
                    "Topic :: Scientific/Engineering :: Bio-Informatics" ]
    )



#!/usr/bin/env python
#-*- coding: utf8
#    Copyright 2013-2015 James S Blachly, MD and The Ohio State University
#
#    This file is part of Mucor.
#
#    Mucor is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    Mucor is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with Mucor.  If not, see <http://www.gnu.org/licenses/>.

#
# mucor_config.py
#
# Karl Kroll
# James S Blachly

import os
import sys
import re
import argparse
from collections import defaultdict
import csv
import json
import codecs
try:
    import HTSeq
except ImportError as err:
    print(str(err) + ". This module is required.")
    sys.exit()

# mucor modules
from inputs import Parser 
from config import Config
import output

def abortWithMessage(message, help = False):
    print("*** FATAL ERROR: " + message + " ***")
    exit(2)

def throwWarning(message, help = False):
    print("\n*** WARNING: " + message + " ***")

def DetectMalformedColumns(fn):
    '''
    Pass a variant calls file, including path, and it returns True if the column values are not consistent with VCF format
    Update: May not be required, now that HTSeq.VCF_Reader is used to parse
    '''
    varFile = open(fn, 'r')
    varReader = csv.reader(varFile, delimiter='\t')
    try:
        row = varReader.next()
    except StopIteration:
        throwWarning("Empty file: {0}".format(fn))
        return True
    while row[0].startswith('##'):
        try:
            row = varReader.next()
        except StopIteration:
            return True
    if row[:9] != ['#CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'FORMAT']:
        return True
    return False


def DetectDataType(fn):
    '''
    Pass a variant calls file, including path, and it will return the name of the tool that created the file. 
    Only works with the expected formats, outlined below.
    Leverages the fact that most variant calling softwares output the name of the tool in every VCF header.
    '''
    unknown_data_type_name = "Unknown" # what to return when the variant caller cannot be inferred from the header
    try:
        varReader = HTSeq.VCF_Reader(str(fn))
        varReader.parse_meta() # extract list of sample names in VCF header
    except IOError:
        throwWarning("File cannot be read, or does not exist: {0}".format(fn))
        return unknown_data_type_name
    except KeyError:
        throwWarning("VCF file cannot be read. Is it malformed? {0}".format(fn))
        return unknown_data_type_name
    # HTSeq.VCF_Reader will accept and attempt to parse existing files that aren't VCF
    #   Parseing completes without warning, but the problem manifests as missing metadata (as if no VCF header)
    meta_info = str(varReader.meta_info()) 
    if not meta_info:
        if os.path.splitext(varReader.fos)[1] == ".vcf" or varReader.fos.endswith(".vcf.gz"):
            throwWarning("VCF file has no metadata. Is this a VCF with a header? {0}".format(fn))
            return unknown_data_type_name
        elif os.path.splitext(varReader.fos)[1] == ".out":
            return "muTector"
    if str('Torrent Unified Variant Caller') in meta_info: 
        return "IonTorrent"
    elif str('MiSeq') in meta_info:
        return "MiSeq"
    elif str('SomaticIndelDetector') in meta_info:
        return "SomaticIndelDetector"
    elif str('MuTect') in meta_info:
        return "Mutect"
    elif str('samtools') in meta_info:
        return "Samtools"
    elif str('VarScan') in meta_info:
        return "VarScan"
    elif str('HaplotypeCaller') in meta_info:
        return "HaplotypeCaller"
    elif str('freeBayes') in meta_info:
        return "FreeBayes"
    elif str('GATK') in meta_info: # this check is last, in case a more specific GATK module was detected above
        return "GenericGATK"
    else:
        return unknown_data_type_name

def DetectSnpEffStatus(fn):
    '''
    Pass a variant calls file, including path, and return True or False according to whether SnpEff has been run on this file.
    '''
    try:
        varReader = HTSeq.VCF_Reader(str(fn))
        varReader.parse_meta() # extract list of sample names in VCF header
    except IOError:
        throwWarning("File cannot be read, or does not exist: {0}".format(fn))
    except KeyError:
        throwWarning("VCF file cannot be read. Is it malformed? {0}".format(fn))
    metadata = varReader.metadata
    if str("SnpEff") in str("metadata"):
        return True
    else:
        return False

def blankJSONDict():
    '''
    Create a blank, valid JSON dictionary to show an example config file. JSON dict keys are identical to the Config class in config.py. 
    '''
    json_dict = defaultdict()
    json_dict['outputDir'] = str("./outputDir_path")
    json_dict['gff'] = str("~/ref/gff_path.gff")
    json_dict['union'] = bool(True)
    json_dict['fast'] = str("~/ref/fastDir_path/") # This will be boolean "False" by default, or a str() if declared
    json_dict['feature'] = str("gene_name")
    json_dict['samples'] = list(dict())

    # VCF filters
    json_dict['filters'] = ["PASS", "lowDP"] 

    # Genomic regions
    json_dict['regions'] = []

    # Variant databases 
    json_dict['databases'] = {"dbName":"~/ref/db_Location.vcf.gz"}

    # Output formats
    json_dict['outputFormats'] = []

    # Samples and associated variant files
    for sid in ['sample1','sample2']:
        # walk over the input project directory (or CWD if none defined) to find all files with sample ID in the name
        tmpSampleDict = defaultdict()
        tmpSampleDict['id'] = str(sid)
        tmpSampleDict['files'] = list()
        for root, dirs, files in ["~/project", "dir1", "sample1-a.vcf"], ["~/project", "dir1", "sample1.bam"], ["~/project", "dir2", "sample1-b.vcf"], ["~/project", "dir3", "sample2.vcf"], ["~/project", "dir3", "sample2.bam"]:
            if str(sid) in str(files): # be careful with sample names here. "U-23" will catch "U-238" etc. Occasional cases can be resolved by manually editing the JSON config file
                full_path = os.path.join(root, dirs, files)
                source = "VarScan"
                if os.path.splitext(files)[1] == str(".vcf") or files.endswith(".vcf.gz"):
                    tmpSampleDict['files'].append({'type':'vcf', 'path':str(full_path), 'snpeff':False, 'source':source} )
                elif os.path.splitext(files)[1] == str(".out"):
                    tmpSampleDict['files'].append({'type':'mutect', 'path':str(full_path), 'source':source} )
                elif os.path.splitext(files)[1] == str(".bam"):
                    try:
                        tmpSampleDict['files'].append({'type':'bam', 'path':str(full_path)} )
                    except KeyError:
                        tmpSampleDict['files'] = list() 
                        tmpSampleDict['files'].append({'type':'bam', 'path':str(full_path)})
            # Not sure if these still work 
                elif os.path.splitext(files)[1].lower() == str(".maf"):
                    tmpSampleDict['files'].append({'type':'maf', 'path':str(full_path), 'source':source} )
                elif os.path.splitext(files)[1].lower() == str(".gvf"):
                    tmpSampleDict['files'].append({'type':'gvf', 'path':str(full_path), 'source':source} )
                else:
                    # If not a VCF, MAF, GVF, or Mutect .out type, ignore it. Uncomment the following line to see the names of files that are being ignored
                    # print("Found an unsupported file type " + str(full_path) + " for sample " + str(sid))
                    pass
                            
        json_dict['samples'].append(tmpSampleDict)
    return json_dict

def processFile(full_path, tmpSampleDict):
    source = DetectDataType(full_path)
    if source == "Unknown" and (os.path.splitext(full_path)[1] == str(".vcf") or full_path.endswith(".vcf.gz") ):
        throwWarning(full_path)
        print("Cannot parse file from an unsupported or unknown variant caller. \nPlease use supported variant software, or compose an input module compatible with inputs.py")
        print("Options include: " + str(Parser().supported_formats.keys()).replace("'",""))
    elif os.path.splitext(full_path)[1] == str(".vcf") or full_path.endswith(".vcf.gz"):
        '''
        # may be handled by HTSeq.VCF_Reader 
        if DetectMalformedColumns(full_path):
            throwWarning(full_path)
            print("Malformed or empty VCF file: {0}".format(full_path))
        '''
        tmpSampleDict['files'].append({'type':'vcf', 'path':str(full_path), 'snpeff':DetectSnpEffStatus(full_path), 'source':source} )
    elif os.path.splitext(full_path)[1] == str(".out"):
        tmpSampleDict['files'].append({'type':'mutect', 'path':str(full_path), 'source':source} )
    elif os.path.splitext(full_path)[1] == str(".bam"):
        try:
            tmpSampleDict['files'].append({'type':'bam','path':str(full_path)} )
            # tmpSampleDict['files'].append({'type':'bam', 'path':str(full_path)} )
        except KeyError:
            tmpSampleDict['files'] = list()
            tmpSampleDict['files'].append({'type':'bam','path':str(full_path)} )
# Not sure if these still work 
    elif os.path.splitext(full_path)[1].lower() == str(".maf"):
        tmpSampleDict['files'].append({'type':'maf', 'path':str(full_path), 'source':source} )
    elif os.path.splitext(full_path)[1].lower() == str(".gvf"):
        tmpSampleDict['files'].append({'type':'gvf', 'path':str(full_path), 'source':source} )
    else:
        # If not a VCF, MAF, GVF, or Mutect .out type, ignore it. Uncomment the following line to see the names of files that are being ignored
        # print("Found an unsupported file type " + str(full_path) + " for sample " + str(sid))
        pass 
    return tmpSampleDict           

def parseAndValidateRegions(regions, json_config=False): 
    '''
    Read in a list of bed file locations and/or string-type chromosome ranges, uniquify it, and return a list of tuples in the form of (chrom, start, stop)
    Note: json_config is an optional argument. It is not used by mucor_config, but will be used by depth_gauge. This is because depth gauge can take regions via the passed json config and from the command line simultaneously. 
    '''
    if json_config:
        unique_region_list = [ x for x in set(regions.split(',')) | set(json_config.regions) if bool(x) ]
        verbose = True 
    else:
        unique_region_list = [ x for x in set(regions.split(',')) if bool(x) ]
        verbose = False # do not print running progress updates if executed from this mucor_config script
    regions_list = []
    if verbose:
        print("== Parsing and Validating Regions ==")
    for i in unique_region_list:
        # the parse
        if os.path.splitext(i)[1].lower() == ".bed":
            # this is a bed file of regions
            # it will be parsed in the main python script
            if os.path.isfile(str(i)):
                regions_list.append(os.path.expanduser(str(i)))
            else:
                throwWarning("BED file {0} cannot be found".format(str(i)))
                continue
        elif str(str(i).split(':')[0]).startswith('chr') and str(str(i).split(':')[0]) != "chr": # these two statements could be combined with a regular expression
            # this looks like a 'chromosome:start-stop' formatted region
            chrom = str(i).split(':')[0]
            start = 0
            end = 0
            try:                                    # does the input region have valid start and ends?
                start = int(re.split('-|\.\.', i.split(':')[1])[0]) # support chr:start-stop and chr:start..stop style location notation
                end   = int(re.split('-|\.\.', i.split(':')[1])[1])
                regions_list.append((str(chrom), start, end))
            except ValueError:                      # start and/or end are invalid, i.e. could not be converted to int
                throwWarning("Region {0} is not valid. Follow standard convention, Ex: chr1:100-300".format(str(i)))
                continue
            except IndexError:                      # only chromosome was defined, or only chromosome:start . this is permitted (whole chromosome region, or single base pair)
                if start and not end:
                    # single point
                    end = start
                    regions_list.append((str(chrom), start, end))
                elif not start and not end:
                    # whole contig
                    regions_list.append((str(chrom), None, None))
        else:
            throwWarning("Region {0} is not a bed file or valid region.".format(i))
        if verbose:
            print("\t{0}".format(i))

    if not regions_list:
        abortWithMessage("Regions not set")
    return regions_list

def getJSONDict(args):
    '''
    Create the JSON dictionary. JSON dict keys are identical to the Config class in config.py. 
    Converts the user input args from argparse into a dictionary
    '''
    json_dict = defaultdict()
    json_dict['outputDir'] = os.path.abspath(os.path.expanduser(str(args['output_directory'])))
    json_dict['gff'] = str(args['gff'])
    json_dict['union'] = bool(args['union'])
    json_dict['fast'] = args['archive_directory']
    json_dict['feature'] = str(args['featuretype'])
    json_dict['samples'] = list(dict())

    # VCF filters
    outFilters = set(["PASS", "."]) # By default, all mutations marked as PASS and '.' are permitted
    for i in str(args['vcf_filters']).split(','):
        if i:
            outFilters.add(i) # Add all user-defined vcf filters. VCFs with empty 'FILTER' columns must have '.' supplied here, or all mutations will be filtered out
    json_dict['filters'] = [x for x in outFilters] # finally, convert set to list

    # Genomic regions
    if args['regions']: # user has defined regions to focus on
        json_dict['regions'] = parseAndValidateRegions(args['regions'])
    else:
        json_dict['regions'] = args['regions']

    # Variant databases 
    json_dict['databases'] = defaultdict(str)
    for i in args['databases']:
        source_name = i.split(':')[0]
        source_path = i.split(':')[1]
        if bool(json_dict['databases'][source_name]):
            abortWithMessage("\ndatabases cannot share the same name.\nrename {0}:{1}\n".format(str(source_name), str(source_path)))
        json_dict['databases'][source_name] = os.path.expanduser(source_path)

    # Output formats
    json_dict['outputFormats'] = []
    for i in str(args['output_type']).split(','):
        json_dict['outputFormats'].append(str(i)) #.lower() was here for some reason

    # Samples and associated variant files
    samples = [] 
    for id in open(args['samples']):
        sid = id.strip()
        if str(sid) == "":
            continue
        else:
            # sample name is not blank line
            samples.append(sid)
    for sid in samples:
        # walk over the input project directory (or CWD if none defined) to find all files with sample ID in the name
        tmpSampleDict = defaultdict()
        tmpSampleDict['id'] = str(sid)
        tmpSampleDict['files'] = list()
        # directory crawl searching for sample name in the file name/path
        if args['project_directory']:
            for root, dirs, files in os.walk(args['project_directory']):
                for i in files:
                    if re.search(r"\b" + re.escape(sid) + r"\b", i): 
                    # be careful with sample names here. "U-23" will catch "U-238" etc. Occasional cases can be resolved by manually editing the JSON config file
                        full_path = os.path.abspath(os.path.join(root, i))
                        tmpSampleDict = processFile(full_path, tmpSampleDict) 
            #json_dict['samples'].append(tmpSampleDict)
        # if files were passed via --inputs
        if args['inputs']:
            for fn in args['inputs']:
                full_path = os.path.abspath(fn)
                if os.path.splitext(fn)[1] == str(".bam"):
                    if re.search(r"\b" + re.escape(sid) + r"\b", fn):
                        tmpSampleDict = processFile(full_path, tmpSampleDict)
                varReader = HTSeq.VCF_Reader(str(fn))
                varReader.parse_meta() # extract list of sample names in VCF header
                if sid in set(varReader.sampleids): 
                    tmpSampleDict = processFile(full_path, tmpSampleDict)
        # uniquify the list of files associated with the sample, in case they were supplied directly via (-i) and found while directory crawling (-d)
        tmpSampleDict['files'] =  { v['path']:v for v in tmpSampleDict['files']}.values()
        json_dict['samples'].append(tmpSampleDict)
    return json_dict

def inconsistentFilesPerSample(json_dict):
    '''
    For each sample, count the number of files found. If any sample has an inconsistent number compared to the others, returns True
    Useful for identifying sample IDs missing an expected file, or sample IDs that erroneously picked up extra files
    '''
    numset = set()
    for sample in json_dict['samples']:
        numset.add(len(sample['files']))
    if len(numset) > 1:
        return True

def filesWithoutSamples(json_dict, inputs):
    '''
    Survey the list of input files (explicitly passed with -i) and return those that are not part of the json dictionary.
    This indicates that the user wanted to analyze them with Mucor, but they were not appropriately identified/associated with a sample ID.
    '''
    # make a list of unique, full file paths associated with each sample
    #   i.e. the files that will be analyzed in the main script
    #   WARNING: complicated, chained python list comprehension ahead! 
    identified_file_paths = set([ os.path.abspath(x['path']) for b in json_dict['samples'] for x in b['files'] ])
    # make a list of unique, full file paths that were input to the config script via (-i)
    input_file_paths =  set([ os.path.abspath(x) for x in inputs ])
    # subtract the sets to identify the files passed via (-i) that were NOT found to be associated with any input sample ID
    missing_file_paths = input_file_paths - identified_file_paths
    return missing_file_paths

class exampleJSON(argparse.Action):
    '''
    This argparse action will run the example JSON config function and exit the program. It is executed when the user gives "-ex || --example" as an argument
    '''
    def __init__(self,option_strings, dest=None, nargs=0, default=None, required=False, type=None, metavar=None, help="Print a valid, example JSON config file and exit."):
        super(exampleJSON, self).__init__(option_strings=option_strings, dest=dest, nargs=nargs, default=default, required=required, metavar=metavar, type=type, help=help)
    def __call__(self, parser, namespace, values, option_string=None):
        print("Printing example_config.json ... ")
        output_file = codecs.open('example_config.json', "w", encoding="utf-8")
        json.dump(blankJSONDict(), output_file, sort_keys=True, indent=4, ensure_ascii=True)
        print("Finished. Exiting ... ")
        sys.exit(0)

def main():
    print
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("-ex", "--example", action=exampleJSON)
    parser.add_argument("-g", "--gff", required=True, help="Annotation GFF/GTF for feature binning")
    parser.add_argument("-f", "--featuretype", required=True, help="Feature type into which to bin. Gencode GTF example: gene_name, gene_id, transcript_name, transcript_id, etc. ")
    parser.add_argument("-db", "--database", dest='databases', metavar='<dbName:/path/database.vcf.gz>', default=[], action='append', help="Colon delimited name and path to variant database in bgzipped VCF format. Can be declared >= 0 times.")
    parser.add_argument("-s", "--samples", metavar='<sample_list.txt>', required=True, help="Text file containing sample names. One sample per line.")
    parser.add_argument("-d", "--project_directory", metavar='<dirname>', required=False, help="Working/project directory, in which to find input variant call files.")
    parser.add_argument("-vcff", "--vcf_filters", default='', help="Comma separated list of VCF filters to allow. Default: PASS") # the defualt value is applied later on in the getJSONDict function, not here.
    parser.add_argument("-a", "--archive_directory", default="", help="Specify directory in which to read/write archived annotations. Undeclared will prevent using the annotation archive features.")
    parser.add_argument("-r", "--regions", default=[], help="Comma separated list of bed regions and/or bed files by which to limit output. Ex: chr1:10230-10240,chr2,my_regions.bed")
    parser.add_argument("-i", "--inputs", nargs="+", help="Input files")
    parser.add_argument("-u", "--union", action="store_true", help="""
        Join all items with same ID for feature_type (specified by -f)
        into a single, continuous bin. For example, if you want intronic
        variants counted with a gene, use this option. 
        WARNING, this will lead to spurious results for features that are duplicated on the same contig.
        When feature names are identical, the bin will range from the beginning of the first instance to the end of the last, even if they are several megabases apart.
        Refer to the documentation for a resolution using 'detect_union_bin_errors.py'
        """)
    parser.add_argument("-jco", "--json_config_output", required=True, help="Name of JSON configuration output file")   
    parser.add_argument("-outd", "--output_directory", required=True, help="Name of directory in which to write Mucor output")
    parser.add_argument("-outt", "--output_type", default="default", help=str("Comma separated list of desired output types. Options include: " + str(output.Writer().supported_formats.keys()).replace("'","") + ". Default: counts,txt"))
    args = vars(parser.parse_args()) # convert the argparser.Namespace to a dictionary, so values can be overwritten 
    
    # Verify user inputs
    # Is there a project directory defined? If so, does it exist?
    if args['project_directory'] and not os.path.exists(args['project_directory']):
        throwWarning("Supplied project directory does not exist: {0}".format(args['project_directory']))
        args['project_directory'] = None
    # Do the intput files exist?
    if args['inputs']:
        for fn in args['inputs']:
            if not os.path.exists(fn):
                throwWarning("Supplied file does not exist: {0}".format(fn))
                # remove this file from inputs
                args['inputs'] = [x for x in args['inputs'] if str(x) != str(fn)]
    # Did the user supply at least 1 valid input file or project directory?
    if not args['project_directory'] and not args['inputs']:
        abortWithMessage("Must supply a valid input file(s) (-i) and/or project directory (-d)")
    # Does the gtf exist?
    if not os.path.exists(args['gff']):
        abortWithMessage("Could not find GFF file {0}".format(args['gff']))

    # Do the database VCFs exist, and are they named properly?
    if args['databases']:
        for db in args['databases']:
            try:
                if not os.path.exists(os.path.expanduser(db.split(':')[1])):
                    abortWithMessage("Could not find SNV DB file {0}".format(db.split(':')[1]))
            except IndexError: # user did not give a name and path, separated by colon
                abortWithMessage("Cannot process {0}\n\tDatabase input must be colon delimited as, 'database_name:database_path'".format(db))

    # Will this output overwrite an existing JSON config file?
    if os.path.exists(args['json_config_output']):
        abortWithMessage("JSON config file {0} already exists.".format(args['json_config_output']))

    # Does the given output directory exist and contain output already?
    if os.path.exists(args['output_directory']) and [x for x in os.listdir(args['output_directory']) if x in output.Writer().file_names.values() ]:
        abortWithMessage("The directory {0} already exists and contains output. Will not overwrite.".format(args['output_directory']))
    elif not os.path.exists(args['output_directory']):
        # If it does not exist, is the parent directory writable?
        if not os.access( os.path.split( os.path.abspath(args['output_directory']) )[0], os.W_OK):
            abortWithMessage("Output directory {0} not writable".format(args['output_directory']))

    # Construct JSON dictionary and dump it to the config file
    json_dict = getJSONDict(args)
    if args['inputs']:
        missing_file_paths = filesWithoutSamples(json_dict, args['inputs']) 
        if missing_file_paths:
            throwWarning("Some input files were not associated with any sample name")
            for fn in missing_file_paths:
                print(fn)
    if inconsistentFilesPerSample(json_dict):
        throwWarning("Inconsistent number of files per sample")
        print("File #\tSample ID")
        for sample in json_dict['samples']:
            print( str(len(sample['files'])) + "\t" + str(sample['id']) )  
    output_file = codecs.open(args['json_config_output'], "w", encoding="utf-8")
    json.dump(json_dict, output_file, sort_keys=True, indent=4, ensure_ascii=True)
    print

if __name__ == "__main__":
	main()
